import { Set, OrderedSet, List, Map, Collection } from "immutable";
export declare class Namespace {
    private readonly _name;
    private readonly _parent?;
    private _children;
    readonly forbiddenNamespaces: Set<Namespace>;
    readonly additionalForbidden: Set<Name>;
    private _members;
    constructor(name: string, parent: Namespace | undefined, forbiddenNamespaces: Set<Namespace>, additionalForbidden: Set<Name>);
    private addChild(child);
    readonly children: OrderedSet<Namespace>;
    readonly members: OrderedSet<Name>;
    readonly forbiddenNameds: Set<Name>;
    add<TName extends Name>(named: TName): TName;
    equals(other: any): boolean;
    hashCode(): number;
}
export declare type NameStyle = (rawName: string) => string;
export declare class Namer {
    readonly name: string;
    readonly nameStyle: NameStyle;
    private readonly _prefixes;
    constructor(name: string, nameStyle: NameStyle, prefixes: string[]);
    assignNames(names: Map<Name, string>, forbiddenNames: Set<string>, namesToAssign: Collection<any, Name>): Map<Name, string>;
}
export declare function funPrefixNamer(name: string, nameStyle: NameStyle): Namer;
export declare abstract class Name {
    private readonly _namingFunction;
    readonly order: number;
    private _associates;
    constructor(_namingFunction: Namer | undefined, order: number);
    equals(other: any): boolean;
    hashCode(): number;
    addAssociate(associate: AssociatedName): void;
    readonly abstract dependencies: List<Name>;
    isFixed(): this is FixedName;
    readonly namingFunction: Namer;
    abstract proposeUnstyledNames(names: Map<Name, string>): OrderedSet<string>;
    firstProposedName: (names: Map<Name, string>) => string;
    nameAssignments(forbiddenNames: Set<string>, assignedName: string): Map<Name, string> | null;
}
export declare class FixedName extends Name {
    private readonly _fixedName;
    constructor(_fixedName: string);
    readonly dependencies: List<Name>;
    addAssociate(_: AssociatedName): never;
    readonly fixedName: string;
    proposeUnstyledNames(_?: Map<Name, string>): OrderedSet<string>;
    hashCode(): number;
}
export declare class SimpleName extends Name {
    private readonly _unstyledNames;
    constructor(_unstyledNames: OrderedSet<string>, namingFunction: Namer, order: number);
    readonly dependencies: List<Name>;
    proposeUnstyledNames(_?: Map<Name, string>): OrderedSet<string>;
    hashCode(): number;
}
export declare class AssociatedName extends Name {
    private readonly _sponsor;
    readonly getName: (sponsorName: string) => string;
    constructor(_sponsor: Name, order: number, getName: (sponsorName: string) => string);
    readonly dependencies: List<Name>;
    proposeUnstyledNames(_?: Map<Name, string>): never;
}
export declare class DependencyName extends Name {
    private readonly _proposeUnstyledName;
    private readonly _dependencies;
    constructor(namingFunction: Namer | undefined, order: number, _proposeUnstyledName: (lookup: (n: Name) => string) => string);
    readonly dependencies: List<Name>;
    proposeUnstyledNames(names: Map<Name, string>): OrderedSet<string>;
    hashCode(): number;
}
export declare function keywordNamespace(name: string, keywords: string[]): Namespace;
export declare function assignNames(rootNamespaces: OrderedSet<Namespace>): Map<Name, string>;
